#line 1 "knap.usr"
/* knapuser.c includes file bb.skl
 *
 * This fork95 program solves the 0-1 Knapsack problem using the
 * Branch and Bound skeleton prototype specified in file bb.skl
 * For information on Branch and Bound and the Knapsack Problem
 * you can use the book of Horowitz and Sahni,
 * "Fundamentals of Computers Algorithms" (Pitman)
 * For more detailed information about the Knapsack problem, see
 * "Knapsack Problems. Algorithms and Computer Implementations." by
 * Silvano Martello and Paolo Toth. John Wiley and Sons.
 * Author: Casiano Rodriguez Leon. Email: casiano@ull.es
 * March, 1999
*/

#include <fork.h>
#include <io.h>
#include <stdlib.h>
#include <assert.h>
#include "tdr.h"
#include "pqueue.h"

/********* using skeleton bb *****************************/
/****** defining input parameters for the skeleton *******/

typedef int k_bb_index;
typedef int k_bb_number;
typedef int boolean;

typedef struct {
	k_bb_number 	k /* object k_bb_index */, 
					C /* remaining capacity */, 
					P /*accumulated profit */;
} k_bb_problemt;

#define MAXNUMBEROFOBJECTS 64 
#define  INFINITY 8192
/* INFINITY has to be set to a value greater 
   than the capacity M of the Knapsack */

#define ns "%d\n"

/* knapsack support variables */
sh k_bb_number N, M;
sh k_bb_number 
		w[MAXNUMBEROFOBJECTS]  /* weights */, 
		p[MAXNUMBEROFOBJECTS]; /* profits */
sh k_bb_number k_bb_bestSol = -INFINITY; 

void readKnap(void) {
#define STRIDE 1

	FILE *in;
  k_bb_index i;

	if ((in = fopen("knap.dat","r+")) == NULL) {
	 fprintf(stderr, "Cannot open input file knap.dat.\n");
	 fprintf(stderr,"Input file format is:\nNumber of objects\nCapacity\nweight 0\nprofit 0\n...\n");
	 exit(1);
	}
	fscanf(in,ns,&N);
	assert(N < MAXNUMBEROFOBJECTS);
	fscanf(in,ns,&M);
	pprintf("N = %d\n",N);
	pprintf("M = %d\n",M);
	assert(INFINITY > M);
	for(i=0;i<N;i++) {
		fscanf(in,ns,&w[i]);
		fscanf(in,ns,&p[i]);
		if (i%STRIDE == 0) {
			pprintf("w = %d\n",w[i]);
			pprintf("p = %d\n",p[i]);
	 }
	}
	w[N] = INFINITY; p[N] = 0; /* sentinel object */
	fclose(in);
}

k_bb_problemt* setKnapProblem(k_bb_number k, k_bb_number C, k_bb_number P) {
	k_bb_problemt *Prob; 

  Prob = (k_bb_problemt *) shmalloc(sizeof(k_bb_problemt)); 

  Prob->k = k; Prob->C = C; Prob->P = P;
	pprintf("setting problem %d %d %d\n",Prob->k,Prob->C,Prob->P);

	return Prob;
}

void k_bb_freeProblem(k_bb_problemt *prob) {

  shfree(prob);
}

void k_bb_lowerUpper(k_bb_problemt *Prob, k_bb_number *L, k_bb_number *U) {
  /* naive approach */
  k_bb_number i, weig = 0, prof = Prob->P;

  if (Prob->C < 0) {
	 *L = -INFINITY; *U = -INFINITY;
  }
  else {
	 for(i=Prob->k; weig <= Prob->C; i++) {
		weig += w[i]; prof += p[i];
	 }
	 /* invariant:  weig > Prob->C */
	 i--;
	 weig -= w[i];
	 prof -= p[i];
	 *L = prof;  /* see Martello and Toth pag 16. Th 2.1 */
	 *U = prof+(p[i]*(Prob->C-weig))/w[i];
  }
}

k_bb_index k_bb_numChildren(k_bb_problemt *Prob) {

  if (Prob->k < N) return 2; else return 0;
}

k_bb_problemt * k_bb_setChild(k_bb_problemt *father,k_bb_index order) {

  return setKnapProblem(father->k+1,
				 father->C-(1-order)*w[father->k],
				 father->P+(1-order)*p[father->k]
				);
}

k_bb_printData(k_bb_problemt Prob) {

	pprintf("(k = %d, C = %d, P = %d)\n",Prob.k,Prob.C,Prob.P);
}
/** The b&b skeleton must be included after the  definition of **/
/********* all the b&b instance functions ***********************/

#include "k_bb.h"

int main(void) {
  k_bb_problemt *Prob;

	if ($ == 0) {
		readKnap();
		Prob = setKnapProblem(0, M, 0);
	}

	pprintf("calling b&b\n...");
	
	k_bb_branchAndBound(Prob);
	
	pprintf("****** best solution = %d *******\n",k_bb_bestSol);
	barrier;
}
